/*
 * iec60958.h:	Define the S/P-DIF interface for redirecting
 *		16bit none audio stream (NO decoding!).
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2002,2003 Werner Fink, <werner@suse.de>
 */
#ifndef __IEC60958_H
#define __IEC60958_H

#include <string.h>
#include "types.h"

#ifdef SPDIF_TEST
# include <stdio.h>
# undef  esyslog
# undef  dsyslog
# define esyslog(format, args...)  fprintf(stderr, format "\n", ## args)
# define dsyslog(format, args...)  fprintf(stderr, format "\n", ## args)
#else
# include <vdr/tools.h>
#endif

#ifndef WORDS_BIGENDIAN
# include <linux/byteorder/swab.h>
# ifndef __swab16
#  ifndef _GNU_SOURCE
extern inline void swab (const void *bfrom, void *bto, ssize_t n)
{
    const char *from = (const char *) bfrom;
    char *to = (char *) bto;

    n &= ~((ssize_t) 1);
    while (n > 1) { // Compiled with -funroll-loops it should be optimized
	const char b0 = from[--n], b1 = from[--n];
	to[n] = b0;
	to[n + 1] = b1;
    }
}
#  endif // if not _GNU_SOURCE
# else	// if __swab16
extern inline void swab (const void *bfrom, void *bto, ssize_t n)
{
    const uint_16 *from = (const uint_16 *)bfrom;
    const uint_16 *end  = from + (n >> 1);
    uint_16 *to = (uint_16 *)bto;

    while (from < end) {
	uint_16 cur = *from++;
	*to++ = __swab16(cur);
    }
}
# endif // if __swab16
#else  // if WORDS_BIGENDIAN
# define swab(bfrom, bto, n)
#endif // if WORDS_BIGENDIAN

typedef enum {
    PCM_STOP   = -1,
    PCM_DATA   =  0,
    PCM_WAIT   =  1,
    PCM_START  =  2,
    PCM_SILENT =  3
} enum_frame_t;

class iec60958 {
protected:
    const unsigned int burst_size;
    typedef frame_t & (*scan_t)(const uint_8 *&, const uint_8 *const);
private:
    //
    // The scanner function for a data section, it sets
    // the current frame hold in `play'
    //
    const scan_t scan;
    const uint_8 offset;
    inline void iec60958::reset(void)
    {
	current  = start;				// At first adress we start
	remember = start + SPDIF_BURST_SIZE;		// Remember the last PCM frame
	payload  = start + offset;			// Pointer to payload of the PCM frame,
							// reserve 4 shorts for lPCM starting head
	pcm      = (uint_16 *)&start[0];		// Provide access to 16bit PCM samples

	memset(start, 0x00, 2*SPDIF_BURST_SIZE);
	play.burst = (uint_32 *)0;
	play.size  = play.pay =  0;
	last.burst = (uint_32 *)0;
	last.size  = last.pay =  0;
    };
public:
    iec60958(const scan_t fscan, unsigned int rate,
	     const unsigned int bsize = SPDIF_BURST_SIZE, const uint_8 poff = 8);
    virtual ~iec60958();
    uint_8 *start;
    uint_8 *current;
    uint_8 *remember;
    uint_8 *payload;
    bool isDVD;
    uint_8 track;
    frame_t play;
    frame_t last;
    uint_16 *pcm;
    void Reset(void);
    inline void Switch(void)
    {
	uint_8 *tmp = current;
	current  = remember;
	remember = tmp;
	payload  = current + offset;
	pcm      = (uint_16 *)&current[0];
    };

    inline void SetBuffer(uint_8 *buf)
    {
	start = buf;					// Buffer with 2*SPDIF_BURST_SIZE bytes
	reset();
    };
    unsigned int sample_rate;
    //
    // Call this in a loop to be sure to catch _all_ encoded frames
    // within out ... end,  returns a pcm frame
    //
    inline const frame_t & Frame(const uint_8 *&out, const uint_8 *const end) const { return scan(out, end); };
    //
    // Returns current pcm frame
    //
    const frame_t & Frame(void) const;
    //
    // Returns various pcm frames (Pause,Start,Stop,Wait) (buffer is remember[])
    //
    const frame_t & Frame(enum_frame_t type);
    inline unsigned int BurstSize (void) const { return B2F(burst_size);  };
    inline unsigned int SampleRate(void) const { return sample_rate; };
    inline void SetErr   (void) const { if (pcm) pcm[2] |=  char2short(0x00, 0x01<<7); };
    inline void ClearErr (void) const { if (pcm) pcm[2] &= ~char2short(0x00, 0x01<<7); };
};

#endif // __IEC60958_H
