/*
 * pts.h:	P(resentation) T(ime) S(tamps) handling
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2003 Werner Fink, <werner@suse.de>
 */

#ifndef _PTS_H
#define _PTS_H

#include <stdint.h>
#include <sys/time.h>
#include "types.h"

enum ePTSstate { eOK = 0, eMISSED = -1, eDOSKIP = 1 };

class cPTS {
private:
    uint_32 start, now;
    struct timeval ts;
public:
    static inline uint_32 PTSinMS(const uchar *pts)
    {
	return (uint_32)
	((   ((((uint64_t)pts[0]) & 0x0E) << 29)
	   | (( (uint64_t)pts[1])         << 22)
	   | ((((uint64_t)pts[2]) & 0xFE) << 14)
	   | (( (uint64_t)pts[3])         <<  7)
	   | ((((uint64_t)pts[4]) & 0xFE) >>  1))/90);
    }

/*

    sender (VDR)                    <->          receiver (spdif)
        |                                            |
     forward credits (ms)            ->           use credits if not expired
   + current time stamp for expire                if no data for credits repeat last frame

                                    <-            if expired put backward credits
                                                  of repeated frames

     if expired get backward credits
     for skipping                    ->           do skipping 
 */

    cPTS(void) : start(0), now(0) {};

    inline void mark(const uchar *pts)
    {
	start = PTSinMS(pts);
    };
    inline ePTSstate compare(const uchar *pts, const uint_32 tolerance)
    {
	mark(pts);
	bool lower = (now - tolerance <= start);
	bool upper = (start <= now + tolerance);
	ePTSstate ret;

	if (upper && lower)
	    ret = eOK;			// OK
	else if (!upper)
	    ret = eMISSED;		// MISSED frames
	else //  !lower
	    ret = eDOSKIP;		// DOSKIP frames
	return ret;
    };
    inline void clear()
    {
	start = 0;
	reset();
    }
    inline void reset()
    {
	now = start;
	gettimeofday(&ts, 0);
    };
    inline void done()
    {
	struct timeval cur, diff;
	gettimeofday(&cur, 0);
	timersub(&cur, &ts, &diff);
	now += (diff.tv_sec*1000+diff.tv_usec/1000);
	gettimeofday(&ts, 0);
    };
    inline uint_32 delay()
    {
	return now - start;
    };
    inline uint_32 getNow() { return now; }

    inline uint_32 getStart() { return start; }

};

#endif

