/*
 * spdif.h:	Access the S/P-DIF interface of sound cards supported by ALSA.
 *		for redirecting none audio streams (NO decoding!).
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * Copyright (C) 2002,2003 Werner Fink, <werner@suse.de>
 */

#include <getopt.h>
#include <unistd.h>
#include "spdif.h"
#include "iec60958.h"

uint_32 spdif::silent_buf[PCM_SILENT_10MS48KHZ];

spdif::spdif(void)
{
    out = NULL;
    stream = NULL;
    delay = 0;
    buffer_size = 0;
    paysize = 0;
    format = SND_PCM_FORMAT_S16_LE;
    (void)snd_pcm_format_set_silence(format, (void*)(&silent_buf[0]), PCM_SILENT_10MS48KHZ);
    silent.burst = &silent_buf[0];
    fragsize = 0;
    period = 0;
    writei = NULL;
    status = NULL;
    log = NULL;
    opt.card = 0;
    opt.device = 2;
    opt.type = SPDIF_CON;
    opt.first = 5;
    opt.mmap = false;
    opt.audio = false;
}

spdif::~spdif()
{
    if (out)
	Close();
}

// (Re)set a stream
bool spdif::Stream(iec60958 *in)
{
    stream = in;
    if (!stream)
	goto xout;
    burst_size = stream->BurstSize();
    periods = (16<<10)/burst_size;
    period = ((burst_size * 1000) / stream->SampleRate());
    switch (stream->SampleRate()) {
    case 48000:
	opt.iec958_aes3_con_fs_rate = IEC958_AES3_CON_FS_48000;
	opt.iec958_aes0_pro_fs_rate = IEC958_AES0_PRO_FS_48000;
	silent.size = (10*48000)/1000;
	break;
    case 44100:
	opt.iec958_aes3_con_fs_rate = IEC958_AES3_CON_FS_44100;
	opt.iec958_aes0_pro_fs_rate = IEC958_AES0_PRO_FS_44100;
	silent.size = (10*44100)/1000;
	break;
    case 32000:
	opt.iec958_aes3_con_fs_rate = IEC958_AES3_CON_FS_32000;
	opt.iec958_aes0_pro_fs_rate = IEC958_AES0_PRO_FS_32000;
	silent.size = (10*32000)/1000;
	break;
    default:
	esyslog("S/P-DIF: Invalid sampling rate (%u)!", stream->SampleRate());
	break;
    }
xout:
    return (stream != NULL);
}

//
// Forward the incoming data to S/P-DIF
//
int spdif::Forward(const uint_8 *const data, const size_t dlen, class cBounce *bounce)
{
    const uint_8 *      head = data;
    const uint_8 *const tail = data + dlen;
    frame_t pcm;
    int count = 0;

    if (!out || !stream)
	goto xout;

    ctrlbits &= ~FL_IO;
    while ((pcm = stream->Frame(head, tail)).burst) {

	if ((ctrlbits & FL_NOEXSYNC) || (ctrlbits & FL_IO))
	    check();
	ctrlbits |= FL_IO;
	count++;

	if (ctrlbits & (FL_FIRST|FL_UNDERRUN|FL_PAUSE|FL_REPEAT|FL_OVERRUN)) {
	    // Let us play with the error detection of the receiver and
	    // send some of the current bursts twice with error bit set.
	    static local int repeat;

	    if (ctrlbits & FL_FIRST) {
		// Time offset
		for (int n = 0; n < opt.first; n++) {
		    burst(silent);	// 10 ms
		    if (check() == SPDIF_HIGH)
			EINTR_RETRY(snd_pcm_wait(out, 10));
		}

		// Start frame
		const frame_t init = (opt.audio) ? stream->Frame(PCM_SILENT) : stream->Frame(PCM_START);
		burst(init);
		ctrlbits &= ~(FL_FIRST|FL_UNDERRUN);

		check();

		repeat = 0;
	    }

	    if (ctrlbits & FL_UNDERRUN) {
		//
		// This recovers the underrun in the case we're getting
		// more input.
		//
		if (ctrlbits & FL_VRPERIOD) {
		    // Let us play with the error detection of the receiver and
		    // send some of the current bursts twice with error bit set.
	    
		    if (!(repeat = (++repeat) % 4)) {
			stream->SetErr();
			burst(pcm);
			stream->ClearErr();
		    }

		    // FL_VRPERIOD will be removed by external thread
		    // checking the buffers state

		} else {
		    // This also recovers the underrun, nevertheless the
		    // stream to the receiver will get short leaks.

		    snd_pcm_drain(out);
		    snd_pcm_prepare(out);
		    ctrlbits &= ~FL_UNDERRUN;
		}

	    } else if (ctrlbits & FL_OVERRUN) {
		//
		// This is a workaround for overruns.
		//
		bool skip = !(bounce->free(buffer_size));

		// Just avoid that burst() is hanging around.
		EINTR_RETRY(snd_pcm_wait(out, -1));
		ctrlbits &= ~FL_OVERRUN;

		// If outer ring buffer is full then skip the burst
		if (skip)
		    continue;
	    }

	    // We play now, no Pause nor we Repeat anymore
	    ctrlbits &= ~(FL_PAUSE|FL_REPEAT);
	}

	burst(pcm);

	// Remember the last pay load size
	paysize = pcm.pay;
    }
xout:
    return count;
}

//
// Put out the burst to S/P-DIF of sound card
//
void spdif::burst(const frame_t &pcm)
{
    snd_pcm_uframes_t frames = B2F(pcm.size);
    const uint_32 *data = pcm.burst;
    int eagain = 0;

    if (!writei)
	goto xout;

    ctrlbits |= FL_BURSTRUN;
    while((frames > 0) && out) {
	snd_pcm_sframes_t res = 0;
	if ((res = writei(out, (const void *)data, frames)) < 0) {
	    switch(res) {
	    case -EAGAIN:
		if (eagain++ > 100) {
			// Real time processes may block
			sync();
			usleep(10*1000);
			goto xout;
		}
		// fall through
	    case -EBUSY:
		EINTR_RETRY(snd_pcm_wait(out, -1));
	    case -EINTR:
		pthread_yield();
		continue;
	    case -EPIPE:
		xunderrun();
		ctrlbits |= FL_FIRST;
		delay = 0;
		break;
	    case -ESTRPIPE:
		xsuspend();
		continue;
	    case -EBADFD:
		if (!out)
		    goto xout;
		(void)snd_pcm_prepare(out);
		continue;
	    default:
		esyslog("S/P-DIF: snd_pcm_writei returned error: %s", snd_strerror(res));
		break;
	    }
	}

	if (res < (snd_pcm_sframes_t)fragsize)
	    EINTR_RETRY(snd_pcm_wait(out, -1));

	frames  -= res;
	data    += res;
    }
xout:
    ctrlbits &= ~FL_BURSTRUN;

    return;
}

//
// Scan options and open the S/P-DIF device
// 
bool spdif::Open(iec60958 *in, ctrl_t &setup)
{
    char pcm_name[256];
    snd_pcm_access_t access;
    unsigned int channels = 2;
    int err, dir;

    if (out)
	Close();
    Lock();

    ctrlbits = FL_FIRST|FL_NOEXSYNC;

    if (!Stream(in))
	goto err_null;

#if 0
    if (!file)
	goto err_null;
    options(file);
    ch.status[0] = IEC958_AES0_NONAUDIO;
#else
    opt.card   = setup.opt.card;
    opt.device = setup.opt.device;
    opt.first  = setup.opt.delay;
    opt.mmap   = setup.opt.mmap;
    if (setup.opt.variable)
	ctrlbits |= FL_VRPERIOD;
    opt.type   = setup.opt.type;
    if (test_setup(AUDIO)) {
	opt.audio = true;
	ch.status[0] = 0;
    } else {
	opt.audio = false;
	ch.status[0] = IEC958_AES0_NONAUDIO;
    }
#endif

    // Note that most alsa sound card drivers uses little endianess
    if (opt.mmap) { 
	writei = snd_pcm_mmap_writei;
	access = SND_PCM_ACCESS_MMAP_INTERLEAVED;
    } else {
	writei = snd_pcm_writei;
	access = SND_PCM_ACCESS_RW_INTERLEAVED;
    }

    switch (opt.type) {
    default:
    case SPDIF_CON:
	ch.status[0] |= (IEC958_AES0_CON_EMPHASIS_NONE);
	ch.status[1]  = (IEC958_AES1_CON_ORIGINAL | IEC958_AES1_CON_PCM_CODER);
	ch.status[2]  =  0;
	ch.status[3]  = (opt.iec958_aes3_con_fs_rate);
	break;
    case SPDIF_PRO:
	ch.status[0] |= (IEC958_AES0_PROFESSIONAL | IEC958_AES0_PRO_EMPHASIS_NONE |
			    opt.iec958_aes0_pro_fs_rate);
	ch.status[1]  = (IEC958_AES1_PRO_MODE_NOTID | IEC958_AES1_PRO_USERBITS_NOTID);
	ch.status[2]  = (IEC958_AES2_PRO_WORDLEN_NOTID);
	ch.status[3]  =  0;
	break;
    }

    if ((err = snprintf(&pcm_name[0], 255, "iec958:AES0=0x%.2x,AES1=0x%.2x,AES2=0x%.2x,AES3=0x%.2x,CARD=%1d",
			ch.status[0], ch.status[1], ch.status[2], ch.status[3], opt.card)) <= 0)
	goto err_null;

    snd_output_stdio_open(&log, "/dev/null", "a");
    if ((err = snd_pcm_open(&out, pcm_name, SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
	// Next try

	snd_pcm_info_t 	*info;
	snd_ctl_elem_value_t *ctl;
	snd_ctl_t *ctl_handle;
	char ctl_name[12];
	int ctl_card;

	if ((err = snprintf(&pcm_name[0], 255, "hw:%1d,%1d", opt.card, opt.device) <= 0))
	    goto err_null;

	snd_output_close(log);
	snd_output_stdio_attach(&log, stderr, 0);
#ifdef DEBUG2
	snd_output_printf(log, "S/P-DIF: Open device %s and try to configure none audio playback ", pcm_name);
	snd_output_printf(log, "(IEC958:AES0=0x%.2x,AES1=0x%.2x,AES2=0x%.2x,AES3=0x%.2x)\n",
			  ch.status[0], ch.status[1], ch.status[2], ch.status[3]);
#endif
	if ((err = snd_pcm_open(&out, pcm_name, SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
	    esyslog("S/P-DIF: sound open: %s", snd_strerror(err));
	    goto err_log;
	}

	snd_pcm_info_alloca(&info);
	if ((err = snd_pcm_info(out, info)) < 0) {
	    esyslog("S/P-DIF: sound info: %s", snd_strerror(err));
	    goto err_out;
	}

	snd_ctl_elem_value_alloca(&ctl);
	snd_ctl_elem_value_set_interface(ctl, SND_CTL_ELEM_IFACE_PCM);
	snd_ctl_elem_value_set_device(ctl, snd_pcm_info_get_device(info));
	snd_ctl_elem_value_set_subdevice(ctl, snd_pcm_info_get_subdevice(info));
	snd_ctl_elem_value_set_name(ctl, SND_CTL_NAME_IEC958("", PLAYBACK,PCM_STREAM));
	snd_ctl_elem_value_set_iec958(ctl, &ch);
	ctl_card = snd_pcm_info_get_card(info);
	if (ctl_card < 0) {
	    esyslog("S/P-DIF: Unable to setup the IEC958 (S/PDIF) interface - PCM has no assigned card");
	    goto __diga_end;
	}
	sprintf(ctl_name, "hw:%d", ctl_card);
	if ((err = snd_ctl_open(&ctl_handle, ctl_name, 0)) < 0) {
	    esyslog("S/P-DIF: Unable to open the control interface '%s': %s", ctl_name, snd_strerror(err));
	    goto __diga_end;
	}
	if ((err = snd_ctl_elem_write(ctl_handle, ctl)) < 0) {
	    esyslog("S/P-DIF: Unable to update the IEC958 control: %s", snd_strerror(err));
	    goto __diga_end;
	}
	snd_ctl_close(ctl_handle);
__diga_end:
	; // make compiler happy
    }
    {
	int fifo = 0;
	unsigned int frag;
	snd_pcm_uframes_t part, period_min, period_max, first;
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_sw_params_t *swparams;

	// Number of digital audio frames in hw buffer
	snd_pcm_uframes_t period_size = 0;
	snd_pcm_sframes_t err;

	snd_pcm_hw_params_alloca(&hwparams);
	snd_pcm_sw_params_alloca(&swparams);

	if ((err = snd_pcm_hw_params_any(out, hwparams)) < 0) {
	    esyslog("S/P-DIF: Broken configuration for this PCM: no configurations available");
	    goto err_out;
	}
	if ((err = snd_pcm_hw_params_set_access(out, hwparams, access)) < 0) {
	    esyslog("S/P-DIF: Access type not available");
	    goto err_out;
	}
	if ((err = snd_pcm_hw_params_set_format(out, hwparams, format)) < 0) {
	    esyslog("S/P-DIF: Sample format not available");
	    goto err_out;
	}

	if ((err = snd_pcm_hw_params_set_channels(out, hwparams, channels)) < 0) {
	    esyslog("S/P-DIF: Channels count not avaible");
	    goto err_out;
	}
#define SND_DIR(dir)	(((dir) > 0) ? ("more") : ("less"))
	dir = 0;
	err = snd_pcm_hw_params_set_rate_near(out, hwparams, stream->SampleRate(), &dir);
	assert(err >= 0);
	if (dir)
	    dsyslog("S/P-DIF: Rate is %s than %d\n", SND_DIR(dir), stream->SampleRate());

#define MMAP_BURST	B2F(getpagesize())
	// Try to use an integer divisor of the burst size as period size
	// to avoid not needed loops and wait states in burst()

	dir = 0;
	period_min = snd_pcm_hw_params_get_period_size_min(hwparams, &dir);
	if (dir)
	    dsyslog("S/P-DIF: Minimal period size is %s than %ld\n", SND_DIR(dir), period_min);
	dir = 0;
	period_max = snd_pcm_hw_params_get_period_size_max(hwparams, &dir);
	if (dir)
	    dsyslog("S/P-DIF: Maximal period size is %s than %ld\n", SND_DIR(dir), period_max);

	if (period_max > burst_size)
	    period_max = burst_size;

	if (opt.mmap)
	    period_max = MMAP_BURST;

	
	frag = 0;
	do {
	    frag++;
	    part = burst_size/frag;

	    if (part > period_max)
		continue;

	    if (part < period_min) {
		err = -ECANCELED;
		break;
	    }

	    if (part*frag != burst_size)
		continue;

	    // function returns less than 0 on error or greater than 0
	    if ((err = snd_pcm_hw_params_set_period_size(out, hwparams, part, 0)) == 0)
		break;

	} while (frag < 10);

	if (err < 0) {
	    esyslog("S/P-DIF: No valid period size available: %s", snd_strerror(err));
	    period_size = 0;
	    goto err_out;
	}

	if (frag > 1) {
	    // function returns less than 0 on error or greater than 0
	    if ((err = snd_pcm_hw_params_set_buffer_size_near(out, hwparams, burst_size*periods)) < 0) {
		esyslog("S/P-DIF: Buffer size not available: %s", snd_strerror(err));
		goto err_out;
	    }
	} else {
	    // function returns less than 0 on error or greater than 0
	    dir = 0;
	    if ((err = snd_pcm_hw_params_set_periods(out, hwparams, periods, dir)) < 0) {
		esyslog("S/P-DIF: Period count not available: %s", snd_strerror(err));
		goto err_out;
	    }
	    if (dir)
		dsyslog("S/P-DIF: Period count %s than %lu\n", SND_DIR(dir), periods);
 	}

	// function returns less than 0 on error or greater than 0
	dir = 0;
	if ((err = snd_pcm_hw_params_get_period_size(hwparams, &dir)) < 0) {
	    esyslog("S/P-DIF: Period size not gotten: %s", snd_strerror(err));
	    goto err_out;
	}
	fragsize = err;
	err *= frag;
	if (frag > 1) {
	    if (dir)
		dsyslog("S/P-DIF: Period size got is %s than %ld\n", SND_DIR(dir), part);
	} else {
	    if (dir)
		dsyslog("S/P-DIF: Burst  size got is %s than %ld\n", SND_DIR(dir), burst_size);
	}
#undef MMAP_BURST
#undef SND_DIR

	if ((int)burst_size != err) {
	    esyslog("S/P-DIF: Period size not set: %s", snd_strerror(err));
	    goto err_out;
	}
	period_size = (snd_pcm_uframes_t)err;
#if defined (SPDIF_SAMPLE_MAGIC) && (SPDIF_SAMPLE_MAGIC > 0)
        fifo = snd_pcm_hw_params_get_fifo_size(hwparams);
	if (fifo <= 0)
	    fifo = SPDIF_SAMPLE_MAGIC;
#endif

	// function returns less than 0 on error or greater than 0
	if ((err = snd_pcm_hw_params_get_buffer_size(hwparams)) < 0) {
	    esyslog("S/P-DIF: Buffer size not set: %s", snd_strerror(err));
	    goto err_out;
	}
	if (period_size * periods != (snd_pcm_uframes_t)err) {
	    esyslog("S/P-DIF: Buffer size not set: %s", snd_strerror(err));
	    goto err_out;
	}
	{
	    snd_pcm_uframes_t tenth;
	    buffer_size = err;
	    tenth = buffer_size/10;
	    buf.upper = buffer_size - 3*tenth;
	    buf.lower = 4*tenth;
	    buf.high  = buffer_size - tenth;
	    buf.alarm = tenth/3;
	}

	if (snd_pcm_hw_params_can_pause(hwparams) == 1)
	    ctrlbits |= FL_CANPAUSE;

	if ((err = snd_pcm_hw_params(out, hwparams)) < 0) {
	    esyslog("S/P-DIF: Cannot set buffer size");
	    snd_pcm_hw_params_dump(hwparams, log);
	    goto err_out;
	}

	if ((err = snd_pcm_sw_params_current(out, swparams)) < 0) {
	    esyslog("S/P-DIF: Cannot get soft parameters: %s", snd_strerror(err));
	    goto err_out;
	}

	// Set start timings
	if ((err = snd_pcm_sw_params_set_sleep_min(out, swparams, 0)) < 0)
	    esyslog("S/P-DIF: Minimal sleep time not available: %s", snd_strerror(err));

	if ((err =  snd_pcm_sw_params_set_xfer_align(out, swparams, fifo)) < 0)
	    esyslog("S/P-DIF: Aligned period size not available: %s", snd_strerror(err));

	if ((err = snd_pcm_sw_params_set_avail_min(out, swparams, period)) < 0)
	    esyslog("S/P-DIF: Minimal period size not available: %s", snd_strerror(err));

	first = silent.size*(opt.first+1);
	if (first > buf.upper)
	    first = buf.upper;

	if ((err = snd_pcm_sw_params_set_start_threshold(out, swparams, first)) < 0)
	    esyslog("S/P-DIF: Start threshold not available: %s", snd_strerror(err));

	if ((err = snd_pcm_sw_params_set_tstamp_mode  (out, swparams, SND_PCM_TSTAMP_MMAP)) < 0)
	    esyslog("S/P-DIF: Time stamp mode not available: %s", snd_strerror(err));

	if ((err = snd_pcm_sw_params(out, swparams)) < 0) {
	    esyslog("S/P-DIF: Cannot set soft parameters: %s", snd_strerror(err));
	    snd_pcm_sw_params_dump(swparams, log);
	    goto err_out;
	}
#ifdef DEBUG2
	snd_pcm_sw_params_dump(swparams, log);
	snd_pcm_dump(out, log);
#endif
    }  

    // Status informations, hold over the full session.
    if ((err = snd_pcm_status_malloc(&status)) < 0) {
	esyslog("S/P-DIF: unable to prepare PCM handle: %s\n", snd_strerror(err));
    }

    if ((err = snd_pcm_prepare(out)) < 0) {
	esyslog("S/P-DIF: unable to prepare PCM handle: %s\n", snd_strerror(err));
	goto err_status;
    }

    Unlock();
    return true;

err_status:
    snd_pcm_status_free(status);
err_out:
    snd_pcm_close(out);
err_log:
    snd_output_close(log);
err_null:
    log = NULL;
    out = NULL;
    status = NULL;
    esyslog("S/P-DIF: unable to establish BitStreamOut for none audio PCM\n");

    Unlock();
    return false;
}

//
// Close the S/P-DIF device
// 
void spdif::Close(void)
{
    bool exit = true;
    snd_pcm_t *tmp = out;

    if (!out)
        return;

    Lock();
    while (ctrlbits & FL_BURSTRUN)
	usleep(1000);
    Clear(exit);

    out    = NULL;
    stream = NULL;

    // Cleanup
    block_signals();
    snd_pcm_nonblock(tmp, SND_PCM_NONBLOCK);
    usleep(1000);
    if (opt.mmap) {
	// Some ALSA version have problems with mmap counter
	// in kernel space, avoid hanging at snd_pcm_close();
	int fd = dup(2); close(2);
	snd_pcm_hw_free(tmp);
	dup2(fd, 2);
	usleep(1000);
    }
    errno = 0;
    snd_pcm_close(tmp);
    leave_signals();

    // Close log
    usleep(1000);
    snd_output_close(log);
    snd_pcm_status_free(status);
    log    = NULL;
    status = NULL;

    Unlock();
}

//
// Clear function, if exit is true send STOP frame
//
void spdif::Clear(bool exit)
{
    snd_pcm_sframes_t err;

    if (!out)
	goto xout;

    ctrlbits |= FL_PAUSE;		// Mute all

    if (ctrlbits & FL_FIRST)
	goto xout;

    if ((err = snd_pcm_status(out, status)) < 0) {
	esyslog("S/P-DIF: clear: status error: %s", snd_strerror(err));
	goto xout;
    }

    switch (snd_pcm_status_get_state(status)) {
    case SND_PCM_STATE_RUNNING:

	if (exit) {
	    ctrlbits &= ~FL_PAUSE;

	    if (stream) {
		// Say decoder to wait for pause or stop
		const frame_t stop = (opt.audio) ? stream->Frame(PCM_SILENT) : stream->Frame(PCM_STOP);
		burst(stop);
	    }

	    if (opt.mmap) {
		// In case of mmap access we've to wait
		delay = snd_pcm_avail_update(out);
		while (delay > 0) {
		    delay = 0;
		    EINTR_RETRY(snd_pcm_wait(out, -1));
		    delay = snd_pcm_avail_update(out);
		}
	    }
	    if ((err = snd_pcm_drop(out)) < 0) {
		switch (err) {
		case -ESTRPIPE:
		    xsuspend();
		    break;
		default:
		    esyslog("S/P-DIF: clear: drop error: %s", snd_strerror(err));
		    break;
		}
	    }
	} else {
	    if ((err = snd_pcm_drain(out)) < 0) {
		switch (err) {
		case -ESTRPIPE:
		    xsuspend();
		    break;
		default:
		    esyslog("S/P-DIF: clear: drain error: %s", snd_strerror(err));
		    break;
		}
	    }
	}
	// fall through
    default:
    case SND_PCM_STATE_XRUN:
	if ((err = snd_pcm_prepare(out)) < 0)
	    esyslog("S/P-DIF: clear: prepare error: %s", snd_strerror(err));
	// fall through
    case SND_PCM_STATE_PREPARED:
	ctrlbits |= FL_FIRST;
	break;
    } // switch (snd_pcm_status_get_state())
xout:
    return;
}

//
// We sleep on PCM stream and watch during this on events on the
// incomming data.
#define SPDIF_REPEAT	320
#define SPDIF_TIMEOUT	(3000 - SPDIF_REPEAT)
bool spdif::Synchronize(class cBounce *bounce)
{
    snd_pcm_sframes_t err;
    bool ready = true;
    int wait;

    if (ctrlbits & FL_FIRST)
	goto do_wait;

    ctrlbits &= ~FL_NOEXSYNC;

repeat:
    if (!out)
	goto do_wait;

    if ((err = snd_pcm_status(out, status)) < 0) {
	esyslog("S/P-DIF: synchronize: status error: %s", snd_strerror(err));
	goto xout;
    }

    switch (snd_pcm_status_get_state(status)) {
    case SND_PCM_STATE_RUNNING:
	switch (check(status)) {
	case SPDIF_HIGH:
	    EINTR_RETRY(snd_pcm_wait(out, -1));
	    check();
	    if (!delay) break;
	    // else fall through
	case SPDIF_OK:
	    wait = (period*(delay-buf.alarm))/burst_size;
	    if (wait < 2*period) {
		if (wait > 0 && bounce->poll(wait))
		    goto xout;
	    } else {
		wait = 2*period;
		if (bounce->poll(wait))
		    goto xout;
		if (xrepeat())		// Repeat last frame for a while
		    goto repeat;
	    }
	default:
	case SPDIF_LOW:
	    if ((err = snd_pcm_drain(out)) < 0) {
		switch (err) {
		case -ESTRPIPE:
		    xsuspend();
		    break;
		default:
		    esyslog("S/P-DIF: synchronize: drain error: %s", snd_strerror(err));
		    break;
		}
	    }
	    break;
	}   // switch(check())
	// fall through
    default:
    case SND_PCM_STATE_XRUN:
	if ((err = snd_pcm_prepare(out)) < 0)
	    esyslog("S/P-DIF: synchronize: prepare error: %s", snd_strerror(err));
	// fall through
    case SND_PCM_STATE_PREPARED:
	ctrlbits |= FL_FIRST;
	goto do_wait;
	break;
    }	// switch (snd_pcm_status_get_state())

xout:
    return ready;

do_wait:
    if (ctrlbits & FL_PAUSE)		// Pause active
	wait = 30;
    else
	wait = SPDIF_TIMEOUT;
    ready = bounce->poll(wait);

    return (ctrlbits & FL_PAUSE) ? true : ready;
}

inline bool spdif::xrepeat(void)
{
    bool ret = true;
    frame_t pcm;

    if (ctrlbits & FL_REPEAT) {
	struct timeval now, diff;
	gettimeofday(&now, 0);
	timersub(&now, &xrstart, &diff);
        ret = ((diff.tv_sec*1000 + diff.tv_usec/1000) < SPDIF_REPEAT);
    } else {
	ctrlbits |= FL_REPEAT;
	gettimeofday(&xrstart, 0);
    }
    if (ret && stream && (pcm = stream->Frame()).burst)
	burst(pcm);

    return ret;
}
#undef SPDIF_REPEAT
#undef SPDIF_TIMEOUT

//
// Internal helper function in case of sound card buffer underrun
//
inline void spdif::xunderrun(void)
{
    snd_pcm_sframes_t res;

    if (!out)
	goto xout;
    if ((res = snd_pcm_status(out, status))<0) {
	esyslog("S/P-DIF: status error: %s", snd_strerror(res));
	goto xout;
    }
    if (snd_pcm_status_get_state(status) == SND_PCM_STATE_XRUN) {
	struct timeval now, diff, tstamp;
	gettimeofday(&now, 0);
	snd_pcm_status_get_trigger_tstamp(status, &tstamp);
	timersub(&now, &tstamp, &diff);
	dsyslog("S/P-DIF: xunderrun!!! (at least %.3f ms long)",
		diff.tv_sec * 1000 + diff.tv_usec / 1000.0);
	if (!out)
	    goto xout;
	if ((res = snd_pcm_prepare(out))<0) {
	    esyslog("S/P-DIF: xunderrun: prepare error: %s", snd_strerror(res));
	    goto xout;
	}
	// ok, data should be accepted again
    }
xout:
    return;
}

//
// Internal helper function in case of sound card buffer suspend
//
inline void spdif::xsuspend (void)
{
    snd_pcm_sframes_t res;

    if (!out)
	goto xout;
    if ((res = snd_pcm_status(out, status))<0) {
	esyslog("S/P-DIF: status error: %s", snd_strerror(res));
	goto xout;
    }
    if (snd_pcm_status_get_state(status) == SND_PCM_STATE_SUSPENDED) {
	esyslog("S/P-DIF: xsuspend!!! trying to resume");
	while (out && (res = snd_pcm_resume(out)) == -EAGAIN)
	    usleep(100000);
	if (!out)
	    goto xout;
	if ((res = snd_pcm_prepare(out))<0) {
	    esyslog("S/P-DIF: xsuspend: prepare error: %s", snd_strerror(res));
	    goto xout;
	}
	// ok, data should be accepted again
    }
xout:
    return;
}

//
// Internal helper to save signal set
//
inline void spdif::block_signals(void)
{
    sigset_t blkset;
    if (sigfillset(&blkset) || sigdelset(&blkset, SIGALRM) ||
	pthread_sigmask(SIG_BLOCK, &blkset, &oldset))
	esyslog("S/P-DIF: can not block signals: %s", strerror(errno));
}

//
// Internal helper to restore signal set
//
inline void spdif::leave_signals(void)
{
    pthread_sigmask(SIG_SETMASK, &oldset, NULL);
}

//
// Internal helper to check the state of the sound card buffer
//
int spdif::check(snd_pcm_status_t * status)
{
    int grade = SPDIF_LOW;
    delay = 0;

    if (ctrlbits & FL_FIRST)	// No AC3 running is LOW
	return grade;

    if (status)
	delay = snd_pcm_status_get_delay(status);
    else
	if (ctrlbits & FL_FIRST)
	    delay = snd_pcm_avail_update(out);
	else
	    snd_pcm_delay(out, &delay);

    if (delay < 0) {
	xunderrun();
	ctrlbits |= FL_FIRST;
	delay = 0;
	goto xout;
    }

    // Underrun dection for setting variable period size
    if      ((snd_pcm_uframes_t)delay <= buf.lower)
	ctrlbits |= FL_UNDERRUN;
    else if ((snd_pcm_uframes_t)delay >  buf.upper)
	ctrlbits &= ~FL_UNDERRUN;

    // Buffer check
    ctrlbits &= ~FL_OVERRUN;
    if      ((snd_pcm_uframes_t)delay <  buf.alarm)
	grade = SPDIF_LOW;
    else if ((snd_pcm_uframes_t)delay >  buf.high) {
	ctrlbits |= FL_OVERRUN;
	grade = SPDIF_HIGH;
    } else
	grade = SPDIF_OK;
xout:
    return grade;
}

//
// Return available space for next PCM frames and avoid
// to block burst() by overrun the sound cards buffer
//
size_t spdif::Available(const size_t max)
{
    ssize_t avail = 0;

    if (!out)
	goto xout;

    if (paysize) {
	switch (check()) {
	default:
	case SPDIF_LOW:
	    if (ctrlbits & FL_FIRST)
		break;
	case SPDIF_OK:			// Fill up to upper boundary (and empty bounce buffer)
	    avail = buffer_size - (delay + buf.high);
	    if (avail > 0) {
		avail = paysize * (avail/burst_size);
		break;
	    }
	    // Fall through
	case SPDIF_HIGH:		// Take less as a frame (use bounce buffer as fallback)
	    avail = paysize>>2;
	    break;
	}
    }

    if (ctrlbits & FL_FIRST) {		// Be able to start without delay
	avail = buffer_size - (silent.size*opt.first + 3*burst_size);
	avail = paysize * (avail/burst_size);
    }

    if (avail <= 0)
	avail = 256;

    if ((size_t)avail > max)
	avail = max;
xout:
    return avail;
}

//
// Set Pause bit
//
void spdif::Pause(const bool onoff)
{
    if (onoff)
	ctrlbits |=  FL_PAUSE;
    // else
    //	will be done in burst()
}

#if 0
//
// Internal helper to scan the configuration file
//
void spdif::options(const char *file)
{
    int rate = stream->SampleRate();
    const struct option long_option[] =
    {
	{"help",	0, NULL,	  'h'},
	{"version",	0, NULL,	  'v'},
	{"device",	1, NULL,	  'D'},
	{"card",	1, NULL,	  'c'},
	{"iec958c",	0, NULL,	  'C'},
	{"iec958p",	0, NULL,	  'P'},
	{"frames",	1, NULL,	  'f'},
	{"rate",	0, NULL,	  'r'},
	{"mmap",	0, NULL,	  'm'},
	{"ioctl",	0, NULL,	  'i'},
	{"block",	0, NULL,	  'b'},
	{"48000",	0, (int*)&rate, 48000},
	{"44100",	0, (int*)&rate, 44100},
	{"44000",	0, (int*)&rate, 44100},
	{"32000",	0, (int*)&rate, 32000},
	{"48",		0, (int*)&rate, 48000},
	{"44",		0, (int*)&rate, 44100},
	{"44.1",	0, (int*)&rate, 44100},
	{"32",		0, (int*)&rate, 32000},
	{NULL,		0, NULL,	  0},
    };

    FILE *fopt = fopen(file, "r");
    if (!fopt)
	return;

    char buf[128], * tmp = buf;
    if (!fgets(buf, sizeof(buf), fopt))
	return;
    fclose(fopt);

    if (tmp[strlen(buf) - 1] == '\n')
	tmp[strlen(buf) - 1] = '\0';

    int c = 0;
    char * abuf[20], ** argv = abuf;
    int    argc = 1;
    for (c = argc; c < 20; c++) {
	// Seperators are not only white spaces but also `:;'
	argv[c] = strsep(&tmp, ":; \t");
	if (argv[c])
	    argc++;
    }
    argv[0] = "spdif class";

    if (*argv[1] != '-')
	return;

    // Reset all opt variables to their initial values because vdr has its
    // own options already scanned.
    optarg = NULL;
    optind = opterr = optopt = 0;
    while ((c = getopt_long(argc, argv, "c:D:f:CPr:imv", long_option, NULL)) > 0) {
	switch (c) {
	case 'D':
	case 'c':
	    if (sscanf(optarg, "%u,%u", &opt.card, &opt.device) != 2) {
		esyslog("S/P-DIF: Wrong card,device arguments (%s)", optarg);
		exit(2);
	    }
	    if (opt.card > (8 - 1) || opt.device > (32 - 1)) { // is 32 correct ?
		esyslog("S/P-DIF: Wrong card,device arguments (%u,%u)", opt.card, opt.device);
		exit(2);
	    }
	    break;
	case 'f':
	    if (sscanf(optarg, "%u", &opt.first) != 1) {
		esyslog("S/P-DIF: Wrong number argument for LPCM 10ms frames (%s)", optarg);
		exit(2);
	    }
	    if (opt.first < 0 || opt.first > 32) {
		esyslog("S/P-DIF: Wrong number argument for LPCM 10ms frames (%u)", opt.first);
		exit(2);
	    }
	    // We use a character buffer for 16bit LPCM
	    break;
	case 'C': opt.type = SPDIF_CON; break;
	case 'P': opt.type = SPDIF_PRO; break;
	case 'r':
	    if (sscanf(optarg, "%u", &rate) != 1) {
		esyslog("S/P-DIF: Wrong number argument for LPCM rate (%s)", optarg);
		exit(2);
	    }
	    break;
	case 'i':
	    opt.mmap = false;
	    break;
	case 'm':
	    opt.mmap = true;
	    break;
	case 'v':
	    ctrlbits |= FL_VRPERIOD;
	    break;
	default: break;
	}
    }

    argv += optind;
    argc -= optind;

    switch (rate) {
    case 48000:
	opt.iec958_aes3_con_fs_rate = IEC958_AES3_CON_FS_48000;
	opt.iec958_aes0_pro_fs_rate = IEC958_AES0_PRO_FS_48000;
	break;
    case 44100:
	opt.iec958_aes3_con_fs_rate = IEC958_AES3_CON_FS_44100;
	opt.iec958_aes0_pro_fs_rate = IEC958_AES0_PRO_FS_44100;
	break;
    case 32000:
	opt.iec958_aes3_con_fs_rate = IEC958_AES3_CON_FS_32000;
	opt.iec958_aes0_pro_fs_rate = IEC958_AES0_PRO_FS_32000;
	break;
    default:
	esyslog("S/P-DIF: Invalid sampling rate (%u)!", rate);
	break;
    }
}
#endif
